from bespin.helpers import memoized_property

import boto.ec2
import boto.ec2.autoscale

import datetime
import logging

log = logging.getLogger("bespin.amazon.ec2")

class EC2(object):
    def __init__(self, region="ap-southeast-2"):
        self.region = region

    @memoized_property
    def conn(self):
        return boto.ec2.connect_to_region(self.region)

    @memoized_property
    def autoscale(self):
        return boto.ec2.autoscale.connect_to_region(self.region)

    def get_instances_in_asg_by_lifecycle_state(self, asg_physical_id, lifecycle_state=None):
        instances = []

        asg = self.autoscale.get_all_groups(names=[asg_physical_id])
        for instance in asg[0].instances:
            if lifecycle_state is None or lifecycle_state == instance.lifecycle_state:
                instances.append(instance.instance_id)

        return instances

    def resume_processes(self, asg_physical_id):
        self.autoscale.suspend_processes(asg_physical_id, ["ScheduledActions"])

    def suspend_processes(self, asg_physical_id):
        self.autoscale.suspend_processes(asg_physical_id, ["ScheduledActions"])

    def display_instances(self, asg_physical_id):
        log.info("Finding instances")
        asg = self.autoscale.get_all_groups(names=[asg_physical_id])
        instance_ids = [inst.instance_id for inst in asg[0].instances]
        print("Found {0} instances".format(len(instance_ids)))
        print("=" * 20)
        if instance_ids:
            for instance in self.conn.get_only_instances(instance_ids=instance_ids):
                launch_time = datetime.datetime.strptime(instance.launch_time, '%Y-%m-%dT%H:%M:%S.000Z')
                delta = (datetime.datetime.utcnow() - launch_time).seconds
                print("{0}\t{1}\t{2}\tUp {3} seconds".format(instance.id, instance.private_ip_address, instance.state, delta))




from bespin.errors import BadOption, BadOptionFormat, MissingVariable
from bespin.option_spec import stack_objs

from input_algorithms.dictobj import dictobj
from option_merge import MergedOptions
from itertools import chain
import string
import six

class Bespin(dictobj):
    fields = [
          "dry_run", "assume_role", "flat", "config", "chosen_stack", "no_assume_role", "extra_imports"
        , "region", "environment", "chosen_artifact", "chosen_task", "extra", "interactive", "configuration"
        ]

    def get_variable(self, artifact):
        try:
            val = self.configuration.root()
            env_objs = []
            for part in artifact.split(','):
                val = val[part]
                if hasattr(val, "env"):
                    env_objs.append(val.env)
                if hasattr(val, "build_env"):
                    env_objs.append(val.build_env)

            if isinstance(val, stack_objs.StaticVariable):
                val = val.resolve()
            elif isinstance(val, stack_objs.DynamicVariable):
                self.set_credentials()
                val = val.resolve()
            if isinstance(val, MergedOptions):
                val = val.as_dict()

            if env_objs and isinstance(val, six.string_types):
                envs = dict(chain.from_iterable([(env.env_name, env) for env in env_obj] for env_obj in env_objs))
                wanted = []
                class Formatter(string.Formatter):
                    def get_field(self, key, args, kwargs, format_spec=None):
                        if key not in envs:
                            raise BadOptionFormat("Couldn't find an environment specification", wanted=key, available=list(envs.keys()))
                        wanted.append(envs[key])

                        if envs[key].missing:
                            return '', key
                        else:
                            return envs[key].pair[1], key
                val = Formatter().format(val)
                missing = [env.env_name for env in wanted if env.missing]
                if any(missing):
                    raise BadOption("Some environment variables aren't in the current environment", missing=missing)

            return val
        except KeyError:
            raise MissingVariable(wanted=artifact)




from bespin.errors import Throttled

from contextlib import contextmanager
import logging
import boto

log = logging.getLogger("bespin.amazon.mixin")

class AmazonMixin(object):
    @contextmanager
    def catch_boto_400(self, errorkls, message, **info):
        """Turn a BotoServerError 400 into a BadAmazon"""
        try:
            yield
        except boto.exception.BotoServerError as error:
            if error.status == 400:
                log.error("%s -(%s)- %s", message, error.code, error.message)
                raise errorkls(message, error_code=error.code, error_message=error.message, **info)
            else:
                raise

    @contextmanager
    def ignore_throttling_error(self):
        try:
            yield
        except boto.exception.BotoServerError as error:
            if error.status == 400 and error.code == "Throttling":
                raise Throttled()
            else:
                raise



from bespin.actions.builder import Builder
from bespin.errors import NoSuchStack

import logging
import time
import json

log = logging.getLogger("bespin.actions.deployer")

class Deployer(object):
    def deploy_stack(self, stack, stacks, made=None, ignore_deps=False, checked=None):
        """Deploy a stack and all it's dependencies"""
        made = [] if made is None else made
        checked = [] if checked is None else checked
        if stack.name in made:
            return

        Builder().sanity_check(stack, stacks, ignore_deps=ignore_deps, checked=checked)
        made.append(stack.name)

        if stack.name not in stacks:
            raise NoSuchStack(looking_for=stack.name, available=stacks.keys())

        if not ignore_deps and not stack.ignore_deps:
            for dependency in stack.dependencies(stacks):
                self.deploy_stack(stacks[dependency], stacks, made=made, ignore_deps=True, checked=checked)

        # Should have all our dependencies now
        log.info("Making stack for '%s' (%s)", stack.name, stack.stack_name)
        self.build_stack(stack)

        if any(stack.build_after):
            for dependency in stack.build_after:
                self.deploy_stack(stacks[dependency], stacks, made=made, ignore_deps=True, checked=checked)

        if stack.artifact_retention_after_deployment:
            Builder().clean_old_artifacts(stack)

        self.confirm_deployment(stack)

    def build_stack(self, stack):
        """Build a single stack"""
        if stack.suspend_actions:
            self.suspend_cloudformation_actions(stack)

        print("Building - {0}".format(stack.stack_name))
        print(json.dumps(stack.params_json_obj, indent=4))

        skip = False
        if stack.cloudformation.status.exists:
            if stack.skip_update_if_equivalent and all(check.resolve() for check in stack.skip_update_if_equivalent):
                log.info("Stack is determined to be the same, not updating")
                skip = True

        changed = False
        if not skip:
            changed = stack.create_or_update()

        if changed:

            # Avoid race condition
            time.sleep(5)

            stack.cloudformation.wait(rollback_is_failure=True)
        else:
            stack.cloudformation.wait()

        stack.cloudformation.reset()

        if stack.suspend_actions:
            self.resume_cloudformation_actions(stack)

    def confirm_deployment(self, stack):
        """Confirm our deployment"""
        stack.check_sns()
        stack.check_url()

    def suspend_cloudformation_actions(self, stack):
        """Suspend the ScheduledActions for an AutoScaling group in the stack"""
        autoscaling_group_id = stack.sns_confirmation.autoscaling_group_id
        asg_physical_id = stack.physical_id_for(autoscaling_group_id)
        stack.ec2.suspend_processes(asg_physical_id)
        log.info("Suspended Processes on AutoScaling Group %s", asg_physical_id)

    def resume_cloudformation_actions(self, stack):
        """Resume the ScheduledActions for an AutoScaling group in the stack"""
        autoscaling_group_id = stack.sns_confirmation.autoscaling_group_id
        asg_physical_id = stack.physical_id_for(autoscaling_group_id)
        stack.ec2.resume_processes(asg_physical_id)
        log.info("Resumed Processes on AutoScaling Group %s", asg_physical_id)




from bespin.helpers import memoized_property
from bespin.errors import BadS3Bucket

from six.moves.urllib.parse import urlparse
from contextlib import contextmanager
from collections import namedtuple
from filechunkio import FileChunkIO

import boto.s3
import humanize
import logging
import boto
import sys
import os

from boto.s3.key import Key

log = logging.getLogger("bespin.amazon.s3")

S3Location = namedtuple("S3Location", ["bucket", "key", "full"])

class S3(object):

    def __init__(self, region="ap-southeast-2"):
        self.region = region

    @memoized_property
    def conn(self):
        return boto.s3.connect_to_region(self.region)

    def s3_location(self, value):
        """Return us an s3 location record type"""
        info = urlparse(value)
        if info.scheme != "s3":
            raise ValueError("S3 location must be a valid s3 url\tgot={0}".format(value))

        bucket = info.netloc
        if not bucket:
            raise ValueError("S3 location must be a valid s3 url\tgot={0}".format(value))

        key = info.path
        return S3Location(bucket, key, value)

    @contextmanager
    def a_multipart_upload(self, bucket, key):
        mp = None
        try:
            mp = bucket.initiate_multipart_upload(key)
            yield mp
        except:
            if mp:
                mp.cancel_upload()
            raise
        else:
            if mp:
                mp.complete_upload()

    def get_bucket(self, bucket_name):
        try:
            return self.conn.get_bucket(bucket_name)
        except boto.exception.S3ResponseError as error:
            if error.status in (404, 403):
                raise BadS3Bucket("Bucket either doesn't exist or isn't available to you", name=bucket_name)
            else:
                raise

    def list_keys_from_s3_path(self, query_path):
        query = self.s3_location(query_path)
        bucket = self.get_bucket(query.bucket)
        return bucket.list(prefix=query.key[1:])

    def delete_key_from_s3(self, key, dry_run):
        if dry_run:
            print("Would delete {0}".format(key.name))
        else:
            print("Deleted {0}".format(key.name))
            key.delete()

    def upload_file_to_s3(self, source_filename, destination_path):
        source_size = os.stat(source_filename).st_size
        if source_size > 30000000:
            self.upload_file_to_s3_in_parts(source_filename, destination_path)
        else:
            self.upload_file_to_s3_as_single(source_filename, destination_path)

    def upload_file_to_s3_in_parts(self, source_filename, destination_path):
        source_file = open(source_filename, 'rb')
        destination_file = self.s3_location(destination_path)

        source = os.path.abspath(source_file.name)
        source_size = os.stat(source).st_size
        log.info("Uploading from %s (%s) to %s in parts", source, humanize.naturalsize(source_size), destination_file.full)

        bucket = self.get_bucket(destination_file.bucket)

        try:
            with self.a_multipart_upload(bucket, destination_file.key) as mp:
                for chunk, offset, length in self.determine_chunks(source_size, min_chunk=5242881):
                    with FileChunkIO(source, 'r', offset=offset, bytes=length) as fp:
                        log.info("Uploading chunk %s (%s)", chunk+1, humanize.naturalsize(length))
                        mp.upload_part_from_file(fp, part_num=chunk+1)

        except boto.exception.S3ResponseError as error:
            if error.status is 403:
                log.error("Seems you are unable to edit this location :(")
                sys.exit(1)
            else:
                raise

        log.info("Finished uploading")

    def determine_chunks(self, total_size, min_chunk=5242881):
        offset = 0
        offsets = []
        while offset < total_size:
            if total_size - offset < min_chunk:
                break
            offsets.append(offset)
            offset += min_chunk
        offsets.append(total_size)
        log.info("Broken up into %s chunks", len(offsets) - 1)

        for index, offset in enumerate(offsets[:-1]):
            nxt = offsets[index+1]
            yield index, offset, nxt - offset

    def upload_file_to_s3_as_single(self, source_filename, destination_path):
        destination_file = self.s3_location(destination_path)

        source = os.path.abspath(source_filename)
        source_size = os.stat(source).st_size
        log.info("Uploading from %s (%s) to %s in one go", source, humanize.naturalsize(source_size), destination_file.full)

        bucket = self.get_bucket(destination_file.bucket)
        key = Key(bucket)
        key.name = destination_file.key
        key.set_contents_from_filename(source_filename)
        key.close()



"""
This is where the mainline sits and is responsible for setting up the logging,
the argument parsing and for starting up Bespin.
"""

from __future__ import print_function

from bespin.errors import BadOption, UserQuit
from bespin.overview import Overview
from bespin import VERSION

from rainbow_logging_handler import RainbowLoggingHandler
from input_algorithms.spec_base import NotSpecified
from delfick_error import DelfickError
import argparse
import logging
import sys
import os

log = logging.getLogger("bespin.executor")

def setup_logging(verbose=False, silent=False, debug=False):
    log = logging.getLogger("")
    handler = RainbowLoggingHandler(sys.stderr)
    handler._column_color['%(asctime)s'] = ('cyan', None, False)
    handler._column_color['%(levelname)-7s'] = ('green', None, False)
    handler._column_color['%(message)s'][logging.INFO] = ('blue', None, False)
    handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)-7s %(name)-15s %(message)s"))
    log.addHandler(handler)
    log.setLevel([logging.INFO, logging.DEBUG][verbose or debug])
    if silent:
        log.setLevel(logging.ERROR)

    logging.getLogger("boto").setLevel([logging.CRITICAL, logging.ERROR][verbose or debug])
    logging.getLogger("requests").setLevel([logging.CRITICAL, logging.ERROR][verbose or debug])
    return handler

class CliParser(object):
    """Knows what argv looks like"""
    def parse_args(self, argv=None):
        """Split the args into <args> -- <extra_args> and run <args> through our argparse.ArgumentParser"""
        if argv is None:
            argv = sys.argv[1:]

        argv = list(argv)
        args = []
        extras = None
        default_task = NotSpecified
        default_stack = NotSpecified
        default_environment = NotSpecified
        default_artifact = NotSpecified

        if argv:
            if not argv[0].startswith("-"):
                default_task = argv[0]
                argv.pop(0)

            if argv and not argv[0].startswith("-"):
                default_environment = argv[0]
                argv.pop(0)

            if argv and not argv[0].startswith("-"):
                default_stack = argv[0]
                argv.pop(0)

            if argv and not argv[0].startswith("-"):
                default_artifact = argv[0]
                argv.pop(0)

        while argv:
            nxt = argv.pop(0)
            if extras is not None:
                extras.append(nxt)
            elif nxt == "--":
                extras = []
            else:
                args.append(nxt)

        other_args = ""
        if extras:
            other_args = " ".join(extras)

        parser = self.make_parser(default_task=default_task, default_stack=default_stack, default_environment=default_environment, default_artifact=default_artifact)
        args = parser.parse_args(args)
        if default_task is not NotSpecified and args.bespin_chosen_task != default_task:
            raise BadOption("Please don't specify task as a positional argument and as a --task option", positional=default_task, kwarg=args.bespin_chosen_task)
        if default_environment is not NotSpecified and args.bespin_environment != default_environment:
            raise BadOption("Please don't specify environment as a positional argument and as a --environment option", positional=default_environment, kwarg=args.bespin_environment)
        if default_stack is not NotSpecified and args.bespin_chosen_stack != default_stack:
            raise BadOption("Please don't specify stack as a positional argument and as a --stack option", positional=default_stack, kwargs=args.bespin_chosen_stack)
        if default_artifact is not NotSpecified and args.bespin_chosen_artifact != default_artifact:
            raise BadOption("Please don't specify artifact as a positional argument and as an --artifact option", positional=default_artifact, kwargs=args.bespin_chosen_artifact)

        return args, other_args

    def make_parser(self, default_task=NotSpecified, default_stack=NotSpecified, default_environment=NotSpecified, default_artifact=NotSpecified):
        parser = argparse.ArgumentParser(description="Opinionated layer around boto")

        logging = parser.add_mutually_exclusive_group()
        logging.add_argument("--verbose"
            , help = "Enable debug logging"
            , action = "store_true"
            )

        logging.add_argument("--silent"
            , help = "Only log errors"
            , action = "store_true"
            )

        logging.add_argument("--debug"
            , help = "Debug logs"
            , action = "store_true"
            )

        opts = {}
        if os.path.exists("./bespin.yml"):
            opts["default"] = "./bespin.yml"
            opts["required"] = False
        else:
            opts["required"] = True

        if "BESPIN_CONFIG" in os.environ:
            opts["default"] = os.environ["BESPIN_CONFIG"]
            del opts["required"]
        parser.add_argument("--bespin-config"
            , help = "The config file specifying what bespin should care about"
            , type = argparse.FileType("r")
            , **opts
            )

        parser.add_argument("--non-interactive"
            , help = "Make this non interactive"
            , dest = "bespin_interactive"
            , action = "store_false"
            )

        parser.add_argument("--flat"
            , help = "Used by the show command"
            , dest = "bespin_flat"
            , action = "store_true"
            )

        parser.add_argument("--dry-run"
            , help = "Should Bespin take any real action or print out what is intends to do"
            , dest = "bespin_dry_run"
            , action = "store_true"
            )

        extra = {"default": "list_tasks"}
        if default_task is not NotSpecified:
            extra["default"] = default_task
        parser.add_argument("--task"
            , help = "The task to run"
            , dest = "bespin_chosen_task"
            , **extra
            )

        extra = {"default": ""}
        if default_environment is not NotSpecified:
            extra["default"] = default_environment
        parser.add_argument("--environment"
            , help = "Specify an environment to play with"
            , dest = "bespin_environment"
            , **extra
            )

        extra = {"default": ""}
        if default_stack is not NotSpecified:
            extra["default"] = default_stack
        parser.add_argument("--stack"
            , help = "Specify a particular stack"
            , dest = "bespin_chosen_stack"
            , **extra
            )

        extra = {"default": ""}
        if default_artifact is not NotSpecified:
            extra["default"] = default_artifact
        parser.add_argument("--artifact"
            , help = "Specify a particular artifact"
            , dest = "bespin_chosen_artifact"
            , **extra
            )

        extra = {"default": False}
        if os.environ.get("NO_ASSUME_ROLE"):
            extra["default"] = True
        parser.add_argument("--no-assume-role"
            , help = "Don't assume role"
            , dest = "bespin_no_assume_role"
            , **extra
            )

        return parser

    def interpret_args(self, argv):
        """Parse argv, do some transformation and return cli_args suitable for Overview"""
        args, extra = CliParser().parse_args(argv)

        cli_args = {"bespin": {}}
        for key, val in sorted(vars(args).items()):
            if key.startswith("bespin_"):
                cli_args["bespin"][key[7:]] = val
            else:
                cli_args[key] = val
        cli_args["bespin"]["extra"] = extra

        return args, cli_args

def set_boto_useragent():
    """Make boto report bespin as the user agent"""
    __import__("boto")
    useragent = sys.modules["boto.connection"].UserAgent
    if "bespin" not in useragent:
        sys.modules["boto.connection"].UserAgent = "{0} bespin/{1}".format(useragent, VERSION)

def main(argv=None):
    try:
        try:
            args, cli_args = CliParser().interpret_args(argv)
            handler = setup_logging(verbose=args.verbose, silent=args.silent, debug=args.debug)
            set_boto_useragent()
            Overview(configuration_file=args.bespin_config.name, logging_handler=handler).start(cli_args)
        except KeyboardInterrupt:
            if CliParser().parse_args(argv)[0].debug:
                raise
            raise UserQuit()
    except DelfickError as error:
        print("")
        print("!" * 80)
        print("Something went wrong! -- {0}".format(error.__class__.__name__))
        print("\t{0}".format(error))
        if CliParser().parse_args(argv)[0].debug:
            raise
        sys.exit(1)

if __name__ == '__main__':
    main()




from bespin.option_spec.specs import many_item_formatted_spec
from bespin.formatter import MergedOptionStringFormatter
from bespin.option_spec.task_objs import Task
from bespin.errors import BadImport

from input_algorithms.dictobj import dictobj
from input_algorithms import spec_base as sb
import imp

class import_spec(many_item_formatted_spec):
    value_name = "Import specification"
    specs = [sb.string_spec(), sb.string_spec()]
    formatter = MergedOptionStringFormatter

    def create_result(self, directory, import_name, meta, val, dividers):
        directory = sb.directory_spec().normalise(meta, directory)
        return Import(directory, import_name)

class Import(dictobj):
    fields = ['directory', 'import_name']

    def do_import(self, bespin, tasks):
        def task_maker(name, description, action=None, **options):
            if not action:
                action = name
            tasks[name] = Task(action, description=description, options=options, label="Project")
            return tasks[name]

        try:
            args = imp.find_module(self.import_name, [self.directory])
        except ImportError as error:
            raise BadImport(directory=self.directory, importing=self.import_name, error=error)

        try:
            module = imp.load_module(self.import_name, *args)
        except SyntaxError:
            raise BadImport(directory=self.directory, importing=self.import_name, error=error)

        if not hasattr(module, "__bespin__"):
            raise BadImport("Extra import had no __bespin__ defined", directory=self.directory, importing=self.import_name)

        module.__bespin__(bespin, task_maker)





from input_algorithms.errors import BadSpec, BadSpecValue
from delfick_error import DelfickError, ProgrammerError

class BespinError(DelfickError): pass

# Explicitly make these errors in this context
BadSpec = BadSpec
BadSpecValue = BadSpecValue
ProgrammerError = ProgrammerError

########################
###   CONFIGURATION
########################

class BadConfiguration(BespinError):
    desc = "Bad configuration"

class BadOptionFormat(BespinError):
    desc = "Bad option format"

class BadOption(BespinError):
    desc = "Bad Option"

class MissingOutput(BespinError):
    desc = "Couldn't find an output"

class MissingFile(BespinError):
    desc = "Couldn't find a file"

class MissingVariable(BespinError):
    desc = "Couldn't find a variable"

class BadFile(BespinError):
    desc = "bad file"

class BadJson(BespinError):
    desc = "Bad json"

class BadYaml(BespinError):
    desc = "Invalid yaml file"

########################
###   OBJECTS
########################

class BadTask(BespinError):
    desc = "Bad task"

class NoSuchStack(BespinError):
    desc = "Couldn't find stack"

class BadStack(BespinError):
    desc = "Bad stack"

class BadDeployment(BespinError):
    desc = "Failed to get all the correct deployment messages"

class InvalidArtifact(BespinError):
    desc = "Chosen artifact is invalid"

########################
###   AMAZON
########################

class BadAmazon(BespinError):
    desc = "Amazon says no"

class BadS3Bucket(BadAmazon):
    desc = "Bad S3 Bucket"

class BadSQSMessage(BadAmazon):
    desc= "Failed to decode message"

class StackDoesntExist(BadAmazon):
    desc = "Missing stack"

########################
###   OTHER
########################

class UserQuit(BespinError):
    desc = "User quit the program"

class StackDepCycle(BespinError):
    desc = "Stack dependency cycle"

class CouldntKill(BespinError):
    desc = "Failed to kill a process"

class BadImport(BespinError):
    desc = "Failed to import"

class Throttled(BespinError):
    desc = "limit rate exceeded"



"""
The functionality itself for each task.

Each task is specified with the ``a_task`` decorator and indicates whether it's
necessary to provide the task with the object containing all the stacks and/or
one specific stack object.
"""

from bespin.actions.deployer import Deployer
from bespin.actions.builder import Builder
import itertools
import logging
import shlex
import json
import os

log = logging.getLogger("bespin.tasks")

available_tasks = {}
class a_task(object):
    """Records a task in the ``available_tasks`` dictionary"""
    def __init__(self, needs_artifact=False, needs_stack=False, needs_stacks=False, needs_credentials=False):
        self.needs_artifact = needs_artifact
        self.needs_stack = needs_stack
        self.needs_stacks = needs_stack or needs_stacks
        self.needs_credentials = needs_credentials

    def __call__(self, func):
        available_tasks[func.__name__] = func
        func.needs_artifact = self.needs_artifact
        func.needs_stack = self.needs_stack
        func.needs_stacks = self.needs_stacks
        func.needs_credentials = self.needs_credentials
        return func

@a_task()
def list_tasks(overview, configuration, tasks, **kwargs):
    """List the available_tasks"""
    print("Available tasks to choose from are:")
    print("Use the --task option to choose one")
    print("")
    keygetter = lambda item: item[1].label
    tasks = sorted(tasks.items(), key=keygetter)
    for label, items in itertools.groupby(tasks, keygetter):
        print("--- {0}".format(label))
        print("----{0}".format("-" * len(label)))
        sorted_tasks = sorted(list(items), key=lambda item: len(item[0]))
        max_length = max(len(name) for name, _ in sorted_tasks)
        for key, task in sorted_tasks:
            print("\t{0}{1} :-: {2}".format(" " * (max_length-len(key)), key, task.description or ""))
        print("")

@a_task(needs_stacks=True)
def show(overview, configuration, stacks, **kwargs):
    """Show what stacks we have"""
    flat = configuration.get("bespin.flat", False)
    only_pushable = configuration.get("bespin.only_pushable", False)

    for index, layer in enumerate(Builder().layered(stacks, only_pushable=only_pushable)):
        if flat:
            for _, stack in layer:
                print(stack.stack_name)
        else:
            print("Layer {0}".format(index))
            for _, stack in layer:
                print("    {0}".format(stack.display_line()))
            print("")

@a_task(needs_stack=True, needs_credentials=True)
def deploy(overview, configuration, stacks, stack, **kwargs):
    """Deploy a particular stack"""
    Deployer().deploy_stack(stack, stacks)

@a_task(needs_stack=True, needs_credentials=True)
def publish_artifacts(overview, configuration, stacks, stack, **kwargs):
    """Build and publish an artifact"""
    Builder().publish_artifacts(stack)

@a_task(needs_stack=True, needs_credentials=True)
def clean_old_artifacts(overview, configuration, stacks, stack, **kwargs):
    """Cleanup old artifacts"""
    Builder().clean_old_artifacts(stack)

@a_task(needs_stack=True, needs_credentials=True)
def confirm_deployment(overview, configuration, stacks, stack, **kwargs):
    """Confirm deployment via SNS notification for each instance and/or url checks"""
    Deployer().confirm_deployment(stack)

@a_task(needs_artifact=True)
def print_variable(overview, configuration, stacks, stack, artifact, **kwargs):
    """Prints out a variable from the stack"""
    print(configuration["bespin"].get_variable(artifact))

@a_task(needs_stack=True, needs_credentials=True)
def suspend_cloudformation_actions(overview, configuration, stacks, stack, **kwargs):
    """Suspends all schedule actions on a cloudformation stack"""
    Deployer().suspend_cloudformation_actions(stack)

@a_task(needs_stack=True, needs_credentials=True)
def resume_cloudformation_actions(overview, configuration, stacks, stack, **kwargs):
    """Resumes all schedule actions on a cloudformation stack"""
    Deployer().resume_cloudformation_actions(stack)

@a_task(needs_stack=True, needs_credentials=True)
def sanity_check(overview, configuration, stacks, stack, **kwargs):
    """Sanity check a stack and it's dependencies"""
    Builder().sanity_check(stack, stacks)
    log.info("All the stacks are sane!")

@a_task(needs_stack=True, needs_credentials=True)
def instances(overview, configuration, stacks, stack, artifact, **kwargs):
    """Find and ssh into instances"""
    if artifact is None:
        asg_physical_id = stack.cloudformation.map_logical_to_physical_resource_id(stack.ssh.autoscaling_group_name)
        stack.ec2.display_instances(asg_physical_id)
    else:
        stack.ssh.ssh_into(artifact, configuration["$@"])

@a_task()
def bastion(overview, configuration, **kwargs):
    """SSH into the bastion"""
    stack = list(configuration["stacks"].keys())[0]
    configuration["stacks"][stack].ssh.ssh_into_bastion(configuration["$@"])

@a_task(needs_credentials=True)
def execute(overview, configuration, **kwargs):
    """Exec a command using assumed credentials"""
    parts = shlex.split(configuration["$@"])
    configuration["bespin"].credentials.verify_creds()
    os.execvpe(parts[0], parts, os.environ)

@a_task(needs_credentials=True, needs_stack=True)
def tail(overview, configuration, stacks, stack, **kwargs):
    """Tail the deployment of a stack"""
    stack.cloudformation.wait()

@a_task(needs_stack=True)
def params(overview, configuration, stacks, stack, **kwargs):
    """Print out the params"""
    stack.find_missing_env()
    print(stack.stack_name)
    print(json.dumps(stack.params_json_obj, indent=4))

@a_task(needs_stack=True, needs_credentials=True)
def outputs(overview, configuration, stacks, stack, **kwargs):
    """Print out the outputs"""
    print(json.dumps(stack.cloudformation.outputs, indent=4))




from bespin.errors import MissingOutput, BadOption, BadStack, BadJson, BespinError
from bespin.errors import StackDoesntExist, BadDeployment
from bespin.helpers import memoized_property
from bespin import helpers as hp

from input_algorithms.spec_base import NotSpecified
from input_algorithms.dictobj import dictobj
import requests
import fnmatch
import logging
import shlex
import json
import stat
import six
import os
import re

log = logging.getLogger("bespin.option_spec.stack_objs")

class Stack(dictobj):
    fields = [
          "bespin", "name", "key_name", "environment", "stack_json", "params_json"
        , "vars", "stack_name", "env", "build_after", "ignore_deps", "artifacts", "build_first"
        , "skip_update_if_equivalent", "tags", "sns_confirmation", "ssh", "build_env"
        , "artifact_retention_after_deployment", "suspend_actions", "url_checker", "params_yaml"
        ]

    def __repr__(self):
        return "<Stack({0})>".format(self.name)

    def physical_id_for(self, autoscaling_group_id):
        return self.cloudformation.map_logical_to_physical_resource_id(autoscaling_group_id)

    def check_url(self):
        environment = dict(env.pair for env in self.env)
        if self.url_checker is not NotSpecified:
            self.url_checker.wait(environment)

    def check_sns(self):
        environment = dict(env.pair for env in self.env)
        if self.sns_confirmation is not NotSpecified:
            self.sns_confirmation.wait(environment, self.ec2, self.sqs, self.cloudformation)

    def dependencies(self, stacks):
        for key_name in self.build_first:
            yield key_name

        for value in self.vars.values():
            if hasattr(value, "stack") and not isinstance(value.stack, six.string_types):
                yield value.stack.key_name

    @property
    def stack_name(self):
        self.find_missing_env()
        environment = dict([env.pair for env in self.env])
        return self._stack_name.format(**environment)

    @stack_name.setter
    def stack_name(self, val):
        self._stack_name = val

    @property
    def build_after(self):
        for stack in self._build_after:
            if isinstance(stack, six.string_types):
                yield stack
            else:
                yield stack.key_name

    @build_after.setter
    def build_after(self, val):
        self._build_after = val

    @property
    def build_first(self):
        for stack in self._build_first:
            if isinstance(stack, six.string_types):
                yield stack
            else:
                yield stack.key_name

    @build_first.setter
    def build_first(self, val):
        self._build_first = val

    def display_line(self):
        return "Stack {0}".format(self.stack_name)

    def find_missing_env(self, key="env"):
        """Find any missing environment variables"""
        missing = [e.env_name for e in getattr(self, key) if e.missing]
        if missing:
            raise BadOption("Some environment variables aren't in the current environment", missing=missing)

    def find_missing_build_env(self):
        self.find_missing_env("build_env")

    @memoized_property
    def cloudformation(self):
        return self.bespin.credentials.cloudformation(self.stack_name)

    @memoized_property
    def ec2(self):
        return self.bespin.credentials.ec2

    @memoized_property
    def sqs(self):
        return self.bespin.credentials.sqs

    @memoized_property
    def s3(self):
        return self.bespin.credentials.s3

    @property
    def stack_json_obj(self):
        return self.stack_json

    @property
    def params_json_obj(self):
        if self.params_json is NotSpecified:
            params = json.dumps(self.params_json)
        else:
            params = json.dumps([{"ParameterKey": key, "ParameterValue": value} for key, value in self.params_yaml.items()])

        environment = dict([env.pair for env in self.env])

        if any(var.needs_credentials for var in self.vars.values()):
            self.bespin.set_credentials()

        for thing in (self.vars.items(), [env.pair for env in self.env]):
            for var, value in thing:
                key = "XXX_{0}_XXX".format(var.upper())
                if key in params:
                    if not isinstance(value, six.string_types):
                        value = value.resolve()
                    params = params.replace(key, value.format(**environment))

        try:
            return json.loads(params)
        except ValueError as error:
            raise BadJson("Couldn't parse the parameters", filename=self.params_json, stack=self.key_name, error=error)

    def create_or_update(self):
        """Create or update the stack, return True if the stack actually changed"""
        log.info("Creating or updating the stack (%s)", self.stack_name)
        status = self.cloudformation.wait()

        if not status.exists:
            log.info("No existing stack, making one now")
            if self.bespin.dry_run:
                log.info("DRYRUN: Would create stack")
            else:
                return self.cloudformation.create(self.stack_json_obj, self.params_json_obj, self.tags.as_dict() or None)
        elif status.complete:
            log.info("Found existing stack, doing an update")
            if self.bespin.dry_run:
                log.info("DRYRUN: Would update stack")
            else:
                return self.cloudformation.update(self.stack_json_obj, self.params_json_obj)
        else:
            raise BadStack("Stack could not be updated", name=self.stack_name, status=status.name)

        return False

    def sanity_check(self):
        self.find_missing_env()
        if all(isinstance(item, six.string_types) for item in (self.params_json, self.params_yaml)):
            raise BadStack("Need either params_json or params_yaml", looking_in=[self.params_json, self.params_yaml])
        if not any(isinstance(item, six.string_types) for item in (self.params_json, self.params_yaml)):
            raise BadStack("Please don't have both params_json and params_yaml")
        matches = re.findall("XXX_[A-Z_]+_XXX", json.dumps(self.params_json_obj))
        if matches:
            raise BadStack("Found placeholders in the generated params file", stack=self.name, found=matches)
        if self.cloudformation.status.failed:
            raise BadStack("Stack is in a failed state, this means it probably has to be deleted first....", stack=self.stack_name)

        with hp.a_temp_file() as fle:
            json.dump(self.stack_json_obj, open(fle.name, "w"))
            self.cloudformation.validate_template(fle.name)

class StaticVariable(dictobj):
    fields = ["value", ("needs_credentials", False)]

    def resolve(self):
        return self.value

class DynamicVariable(dictobj):
    fields = ["stack", "output", ("bespin", None), ("needs_credentials", True)]

    def resolve(self):
        if isinstance(self.stack, six.string_types):
            cloudformation = self.bespin.credentials.cloudformation(self.stack)
            cloudformation.wait()
            outputs = cloudformation.outputs
        else:
            outputs = self.stack.cloudformation.outputs

        if self.output not in outputs:
            raise MissingOutput(wanted=self.output, available=outputs.keys())

        return outputs[self.output]

class Environment(dictobj):
    """A single environment variable, and it's default or set value"""
    fields = ["env_name", ("default_val", None), ("set_val", None)]

    @property
    def missing(self):
        return self.default_val is None and self.set_val is None and self.env_name not in os.environ

    @property
    def pair(self):
        """Get the name and value for this environment variable"""
        if self.set_val is not None:
            return self.env_name, self.set_val
        elif self.default_val is not None:
            return self.env_name, os.environ.get(self.env_name, self.default_val)
        else:
            return self.env_name, os.environ[self.env_name]

class Skipper(dictobj):
    fields = ["var1", "var2"]

    def resolve(self):
        try:
            v1 = self.var1().resolve()
        except StackDoesntExist:
            return False

        try:
            v2 = self.var2().resolve()
        except StackDoesntExist:
            return False

        return v1 and v2 and v1 == v2

class SSH(dictobj):
    fields = [
          "user", "bastion", "bastion_key_location"
        , "instance_key_location", "autoscaling_group_name"
        , "instance_key_path", "bastion_key_path"
        ]

    def ssh_into_bastion(self, extra_args):
        if not os.path.exists(self.bastion_key_path):
            log.error("Didn't find a bastion key, please download the key")
            print("Bastion key can be found at {0}".format(self.bastion_key_location))
            print("Download it to {0}".format(self.bastion_key_path))
            raise BespinError("Couldn't find an ssh key for the bastion")

        os.chmod(self.bastion_key_path, 0)
        os.chmod(self.bastion_key_path, stat.S_IRUSR)

        command = "ssh {0}@{1} -i {2} -o IdentitiesOnly=true".format(self.user, self.bastion, self.bastion_key_path)
        parts = shlex.split(command)
        os.execvp(parts[0], parts)

    def ssh_into(self, ip_address, extra_args):
        proxy = ""
        error = False
        if self.bastion is not NotSpecified:
            log.info("Logging into %s via %s", ip_address, self.bastion)
            if not os.path.exists(self.bastion_key_path):
                log.error("Didn't find a bastion key, please download the key")
                print("Bastion key can be found at {0}".format(self.bastion_key_location))
                print("Download it to {0}".format(self.bastion_key_path))
                error = True
            proxy = '-o ProxyCommand="ssh {0}@{1} -W %h:%p -i {2} -o IdentitiesOnly=true"'.format(self.user, self.bastion, self.bastion_key_path)
        else:
            log.info("Logging into %s", ip_address)

        if not os.path.exists(self.instance_key_path):
            log.error("Didn't find a instance key, please download the key")
            print("Instance key can be found at {0}".format(self.instance_key_location))
            print("Download it to {0}".format(self.instance_key_path))
            error = True

        if error:
            raise BespinError("Couldn't find ssh keys")

        os.chmod(self.instance_key_path, 0)
        os.chmod(self.instance_key_path, stat.S_IRUSR)
        os.chmod(self.bastion_key_path, 0)
        os.chmod(self.bastion_key_path, stat.S_IRUSR)

        command = "ssh -o ForwardAgent=false -o IdentitiesOnly=true {0} -i {1} {2}@{3} {4}".format(proxy, self.instance_key_path, self.user, ip_address, extra_args)
        parts = shlex.split(command)
        log.debug("Running %s", command)
        os.execvp(parts[0], parts)

class UrlChecker(dictobj):
    fields = ["check_url", "endpoint", "expect", "timeout_after"]

    def wait(self, environment):
        endpoint = self.endpoint().resolve()
        while endpoint.endswith("/"):
            endpoint = endpoint[:-1]
        while endpoint.endswith("."):
            endpoint = endpoint[:-1]

        while self.check_url.startswith("/"):
            self.check_url = self.check_url[1:]

        url = endpoint + '/' + self.check_url
        expected = self.expect.format(**environment)

        log.info("Asking server for version till we match %s", expected)
        for _ in hp.until(self.timeout_after, step=15):
            log.info("Asking %s", url)
            try:
                result = requests.get(url).text
            except requests.exceptions.ConnectionError as error:
                log.warning("Failed to ask server\terror=%s", error)
            else:
                log.info("\tgot back %s", result)
                if fnmatch.fnmatch(result, expected):
                    log.info("Deployment successful!")
                    return

        raise BadStack("Timedout waiting for the app to give back the correct version")

class SNSConfirmation(dictobj):
    fields = ["version_message", "env", "deployment_queue", "autoscaling_group_id", ("timeout", 300)]

    def wait(self, environment, ec2, sqs, cloudformation):
        autoscaling_group_id = self.autoscaling_group_id
        asg_physical_id = cloudformation.map_logical_to_physical_resource_id(autoscaling_group_id)
        instances_to_check = ec2.get_instances_in_asg_by_lifecycle_state(asg_physical_id, lifecycle_state="InService")

        version_message = self.version_message.format(**environment)

        failed = []
        success = []
        attempt = 0

        log.info("Checking sqs for %s", version_message)
        for _ in hp.until(timeout=self.timeout, step=5, action="Checking for valid deployment actions"):
            messages = sqs.get_all_deployment_messages(self.deployment_queue)

            # Look for success and failure in the messages
            for message in messages:
                log.info("Message received %s", message['output'])

                # Ignore the messages for instances outside this deployment
                if message['instance_id'] in instances_to_check:
                    if fnmatch.fnmatch(message['output'], version_message):
                        log.info("Deployed instance %s", message['instance_id'])
                        success.append(message['instance_id'])
                    else:
                        log.info("Failed to deploy instance %s", message['instance_id'])
                        log.info("Failure Message: %s", message['output'])
                        failed.append(message['instance_id'])

            # Stop trying if we have all the instances
            if set(failed + success) == set(instances_to_check):
                break

            # Record the iteration of checking for a valid deployment
            attempt += 1
            log.info("Completed attempt %s of checking for a valid deployment state", attempt)

        if success:
            log.info("Succeeded to deploy %s", success)
        if failed:
            log.error("Failed to deploy %s", failed)
            raise BadDeployment(failed=failed)

        if not success and not failed:
            log.error("Failed to receive any messages")
            raise BadDeployment("Failed to receive any messages")

        log.info("All instances have been confirmed to be deployed with version_message [%s]!", version_message)




from bespin.helpers import memoized_property
from bespin.errors import BadSQSMessage

import boto.sqs

import logging
import json
import time

log = logging.getLogger("bespin.amazon.sqs")

class SQS(object):
    def __init__(self, region="ap-southeast-2"):
        self.region = region

    @memoized_property
    def conn(self):
        return boto.sqs.connect_to_region(self.region)

    def get_all_deployment_messages(self, sqs_url, timeout=60, sleep=2):
        """Get all the messages of the queue, dropping those that are not deployment messages"""
        start = time.time()
        messages = []

        q = self.conn.get_queue(sqs_url)
        while q.count() > 0 and not (time.time() - start > timeout):
            raw_messages = self.conn.receive_message(q, number_messages=1)

            if len(raw_messages) > 0:
                raw_message = raw_messages[0]
                encoded_message = json.loads(raw_message.get_body())['Message']

                q.delete_message(raw_message)

                message = self.decode_message(encoded_message)

                if message is not None:
                    messages.append(message)

            time.sleep(sleep)

        return messages

    def decode_message(self, encoded_message):
        """Takes a message that is : separated and maps it to a Dict"""
        if encoded_message.count(':') < 2:
            raise BadSQSMessage("Less than two colons", msg=encoded_message)

        result, instance_id, output = encoded_message.split(':', 2)
        return {'result': result, 'instance_id': instance_id, 'output': output }

